-- Setting constants
n = 0
cyclic = false
dfa = true

-- Reading parameter and flags

if(#arg == 0 or arg[1] == '-h' or arg[1] == '--help') then
    print('Usage: lua generate-aut.lua [OPTIONS] n for generating the DBA/UBA of the n-last-bit-example')
    print('Options:')
    print('--uba        for generating an UBA instead of DBA')
    print('--cyclic     for generating the cyclic version')
    return
end

if(arg[1] == '--cyclic') then
    cyclic = true
end
if(arg[1] == '--uba') then
    dfa = false
end
if(string.find(arg[1], '%d+') ~= nil) then
    n = tonumber(string.match(arg[1], '(%d+)')) 
end
if(arg[2] ~= nil) then
    if(arg[2] == '--cyclic') then
        cyclic = true
    end
    if(arg[2] == '--uba') then
        dfa = false
    end
    if(string.find(arg[2], '%d+') ~= nil) then
        n = tonumber(string.match(arg[2], '(%d+)')) 
    end
end
if(arg[3] ~= nil) then
    if(arg[3] == '--cyclic') then
        cyclic = true
    end
    if(arg[3] == '--uba') then
        dfa = false
    end
    if(string.find(arg[3], '%d+') ~= nil) then
        n = tonumber(string.match(arg[3], '(%d+)')) 
    end
end

-- Help functions: 2 to the power of as integer
myPow = function(exponent)
    return math.floor(2^exponent)
end

-- Shifting help functions
shiftLeft = function(number,positions)
    return math.floor(myPow(positions)*number % myPow(n+1))
end

bitAt = function(number, position)
    return math.floor(number / myPow(position - 1)) % 2
end


-- generate (cyclic) UBA for given n (in description k)
generateUBA = function(n)
    local ufa = 'HOA: v1\nStates: ' .. n+2 .. '\nStart: 0\nAP: 2 "L0" "L1"\nAcceptance: 1 Inf(0)\nacc-name: Buchi\nproperties: unambiguous state-acc trans-labels\n--BODY--\n'
    ufa = ufa .. 'State: 0\n[!0] 0\n[!0 & 1] 1\n'
    for i = 1,n-1 do
        ufa = ufa .. 'State: ' .. i .. '\n' .. '[!0] ' .. i + 1 .. '\n'
    end
    ufa = ufa .. 'State: ' .. n  .. '\n' .. '[0] ' .. n + 1 .. '\n'
    if(cyclic) then
        ufa = ufa .. 'State: ' .. n + 1  .. '{0}\n' .. '[!0] ' .. 0 .. '\n'
    else
        ufa = ufa .. 'State: ' .. n + 1 .. '{0}\n' .. '[t] ' .. n + 1 .. '\n'
    end
    ufa = ufa .. '--END--'
    return ufa
end 


-- generate DBA
generateDBA = function(n) 
    -- HOA Header
    local dba = 'HOA: v1\nStates: ' .. myPow(n+1) + 1 .. '\nStart: 0\nAP: 2 "L0" "L1"\nAcceptance: 1 Inf(0)\nacc-name: Buchi\nproperties: deterministic state-acc trans-labels\n--BODY--\n'
    for i = 0,myPow(n+1) - 1 do 
        if i < myPow(n+1) - 1 then
            dba = dba .. 'State: ' .. i .. '\n'
            -- no error, no condition satisfied
            dba = dba .. succToHOA(i, false, false)
            -- error occured, but no condition satisfied
            dba = dba .. succToHOA(i, true, false)
            -- no error occurred, but condition satisfied
            dba = dba .. succToHOA(i, false, true)
            -- error occurred and, condition satisfied
            dba = dba .. succToHOA(i, true, true)
        else
            if(cyclic) then
                -- go back to initial state
                dba = dba .. 'State: ' .. i .. ' {0}\n[!0] 0\n[0] ' .. myPow(n+1) .. '\n'
            else
                -- stay in initial state
                dba = dba .. 'State: ' .. i .. ' {0}\n[t] ' .. i .. '\n'
            end
        end
    end
    dba = dba .. 'State: ' .. myPow(n+1) .. '\n[t] ' .. myPow(n+1) .. '\n'
    dba = dba .. '--END--'
    return dba
end

-- successor for one HOA state in DBA
succToHOA = function(index, error, condition)
    local succ,valid = generateDBASuccessor(index,error,condition)
    succ = math.floor(succ)
    -- write symbol (error, condition)
    local result = '['
    if (error) then
        result = result .. '0'
    else
        result = result .. '!0'
    end
    result = result .. ' & '
    if (condition) then
        result = result .. '1'
    else
        result = result .. '!1'
    end
    result = result .. '] '
    if(valid) then
        result = result .. succ
    else
        result = result .. myPow(n+1)
    end
    result = result .. '\n'
    return result
end

-- Generate the successor index
generateDBASuccessor = function(index,error,condition) 
    if(index == myPow(n) - 1 and not error) then
        if(condition) then
            return myPow(n) - 1,true
        else
            return myPow(n) - 2,true
        end
    end

    if(index == myPow(n+1) -1) then
        if(not cyclic) then
            return index,true
        else
            if(not error) then
                return 0,true
            end
        end
    end

    if(not error) then
        newState = shiftLeft(index, 1)
        if(condition) then
            newState = newState + 1
        end
        return newState,true
    else
        local bitAtN = math.floor((index / myPow(n-1))) % 2
        if(bitAtN >= 1) then
            return myPow(n+1)-1,true
        else
            return 0,false
        end
    end
end

if(dfa) then
    print(generateDBA(tonumber(n)))
else
    print(generateUBA(tonumber(n)))
end
